@@ -115,8 +115,8 @@ zgetcwd(void)
     struct stat sbuf;
     struct dirent *de;
     DIR *dir;
-    ino_t ino, rootino = (ino_t) ~ 0;
-    dev_t dev, rootdev = (dev_t) ~ 0;
+    ino_t ino, pino, rootino = (ino_t) ~ 0;
+    dev_t dev, pdev, rootdev = (dev_t) ~ 0;
 
     holdintr();
     buf2[0] = '\0';
@@ -127,20 +127,27 @@ zgetcwd(void)
 	rootdev = sbuf.st_dev;
     }
 
+    if (stat(".", &sbuf) < 0) {
+	noholdintr();
+	return ztrdup(".");
+    }
+
+    pino = sbuf.st_ino;
+    pdev = sbuf.st_dev;
+
     for (;;) {
-	if (stat(".", &sbuf) < 0) {
-	    chdir(buf0);
-	    noholdintr();
-	    return ztrdup(".");
-	}
-	ino = sbuf.st_ino;
-	dev = sbuf.st_dev;
 	if (stat("..", &sbuf) < 0) {
 	    chdir(buf0);
 	    noholdintr();
 	    return ztrdup(".");
 	}
-	if ((sbuf.st_ino == ino && sbuf.st_dev == dev) ||
+
+	ino = pino;
+	dev = pdev;
+	pino = sbuf.st_ino;
+	pdev = sbuf.st_dev;
+
+	if ((ino == pino && dev == pdev) ||
 	    (ino == rootino && dev == rootdev)) {
 	    chdir(buf0);
 	    noholdintr();
@@ -160,35 +167,23 @@ zgetcwd(void)
 		(fn[1] == '\0' ||
 		 (fn[1] == '.' && fn[2] == '\0')))
 		continue;
-	    if ((ino_t) de->d_ino == ino) {
+	    if (dev != pdev || (ino_t) de->d_ino == ino) {
 		lstat(fn, &sbuf);
-		if (sbuf.st_dev == dev)
-		    goto match;
+		if (sbuf.st_dev == dev && sbuf.st_ino == ino) {
+		    strcpy(buf3, de->d_name);
+		    break;
+		}
 	    }
 	}
 	closedir(dir);
-	dir = opendir(".");
-	while ((de = readdir(dir))) {
-	    char *fn = de->d_name;
-	    /* Ignore `.' and `..'. */
-	    if (fn[0] == '.' &&
-		(fn[1] == '\0' ||
-		 (fn[1] == '.' && fn[2] == '\0')))
-		continue;
-	    lstat(fn, &sbuf);
-	    if (sbuf.st_dev == dev)
-		goto match;
+	if (!de) {
+	    noholdintr();
+	    return ztrdup(".");
 	}
-	noholdintr();
-	closedir(dir);
-	return ztrdup(".");
-      match:
-	strcpy(buf3, de->d_name);
 	if (*buf2)
 	    strcat(buf3, "/");
 	strcat(buf3, buf2);
 	strcpy(buf2, buf3);
-	closedir(dir);
     }
 }
 
