@@ -97,7 +97,11 @@ evalcond(Cond c)
     case 'w':
 	return (doaccess(c->left, W_OK));
     case 'x':
-	return (doaccess(c->left, X_OK));
+	if (!geteuid()) {
+	    unsigned short mode = dostat(c->left);
+	    return (mode & 0111) || S_ISDIR(mode);
+	}
+	return doaccess(c->left, X_OK);
     case 'z':
 	return (!strlen(c->left));
     case 'h':
