@@ -87,7 +87,7 @@ static struct qual *quals;
 /* Other state values for current pattern */
 static int qualct, qualorct;
 static int range, amc, units;
-static int gf_nullglob, gf_markdirs, gf_noglobdots, gf_listtypes;
+static int gf_nullglob, gf_markdirs, gf_noglobdots, gf_listtypes, gf_follow;
 
 /* Prefix, suffix for doing zle trickery */
 char *glob_pre, *glob_suf;
@@ -157,7 +157,7 @@ glob(LinkList list, LinkNode np)
     colonmod = NULL;
     gf_nullglob = isset(NULLGLOB);
     gf_markdirs = isset(MARKDIRS);
-    gf_listtypes = 0;
+    gf_listtypes = gf_follow = 0;
     gf_noglobdots = unset(GLOBDOTS);
     if (str[sl - 1] == Outpar) {	/* check for qualifiers */
 	char *s;
@@ -411,11 +411,13 @@ glob(LinkList list, LinkNode np)
 			break;
 		    case 'M':
 			/* Mark directories with a / */
-			gf_markdirs = !(sense & 1);
+			if ((gf_markdirs = !(sense & 1)))
+			    gf_follow = sense & 2;
 			break;
 		    case 'T':
 			/* Mark types in a `ls -F' type fashion */
-			gf_listtypes = !(sense & 1);
+			if ((gf_listtypes = !(sense & 1)))
+			    gf_follow = sense & 2;
 			break;
 		    case 'N':
 			/* Nullglob:  remove unmatched patterns. */
@@ -637,20 +639,29 @@ void
 insert(char *s)
 {
     struct stat buf, buf2, *bp;
+    char *news = s;
     int statted = 0;
 
     if (gf_listtypes || gf_markdirs) {
 	/* Add the type marker to the end of the filename */
-	statted = 1;
-	if (!lstat(unmeta(s), &buf) && (gf_listtypes || S_ISDIR(buf.st_mode))) {
-	    char *t;
-	    int ll = strlen(s);
+	statted = -1;
+	if (!lstat(unmeta(s), &buf)) {
+	    mode_t mode = buf.st_mode;
+	    statted = 1;
+	    if (gf_follow) {
+		if (!S_ISLNK(mode) || stat(unmeta(s), &buf2))
+		    memcpy(&buf2, &buf, sizeof(buf));
+		statted = 2;
+		mode = buf2.st_mode;
+	    }
+	    if (gf_listtypes || S_ISDIR(mode)) {
+		int ll = strlen(s);
 
-	    t = (char *)ncalloc(ll + 2);
-	    strcpy(t, s);
-	    t[ll] = file_type(buf.st_mode);
-	    t[ll + 1] = '\0';
-	    s = t;
+		news = (char *)ncalloc(ll + 2);
+		strcpy(news, s);
+		news[ll] = file_type(buf.st_mode);
+		news[ll + 1] = '\0';
+	    }
 	}
     }
     if (qualct || qualorct) {
@@ -658,8 +669,7 @@ insert(char *s)
 	struct qual *qo, *qn;
 	int t = 0;		/* reject file unless t is set */
 
-	if (statted || lstat(unmeta(s), &buf) >= 0) {
-	    statted = 0;
+	if (statted >= 0 && (statted || !lstat(unmeta(s), &buf))) {
 	    for (qo = quals; qo && !t; qo = qo->or) {
 
 		t = 1;
@@ -667,17 +677,17 @@ insert(char *s)
 		    range = qn->range;
 		    amc = qn->amc;
 		    units = qn->units;
-		    if ((qn->sense & 2) && !statted) {
+		    if ((qn->sense & 2) && statted != 2) {
 			/* If (sense & 2), we're following links */
-			statted = 1;
-			stat(unmeta(s), &buf2);
+			if (!S_ISLNK(buf.st_mode) || stat(unmeta(s), &buf2))
+			    memcpy(&buf2, &buf, sizeof(buf));
+			statted = 2;
 		    }
 		    bp = (qn->sense & 2) ? &buf2 : &buf;
 		    /* Reject the file if the function returned zero *
-		     * and the sense was positive (sense == 0), or   *
+		     * and the sense was positive (sense&1 == 0), or *
 		     * vice versa.                                   */
-		    if (!(!!((qn->func) (bp, qn->data)) ^
-			  (qn->sense & 1))) {
+		    if ((!((qn->func) (bp, qn->data)) ^ qn->sense) & 1) {
 			t = 0;
 			break;
 		    }
@@ -689,11 +699,10 @@ insert(char *s)
     }
     if (colonmod) {
 	/* Handle the remainder of the qualifer:  e.g. (:r:s/foo/bar/). */
-	char *cm2 = colonmod;
-
-	modify(&s, &cm2);
+	s = colonmod;
+	modify(&news, &s);
     }
-    *matchptr++ = s;
+    *matchptr++ = news;
     if (++matchct == matchsz) {
 	matchbuf = (char **)realloc((char *)matchbuf,
 				    sizeof(char **) * (matchsz *= 2));
@@ -1442,7 +1451,7 @@ static int addpath _((char *s));
 static int
 addpath(char *s)
 {
-    if ((int)strlen(s) + pathpos >= PATH_MAX)
+    if ((int)strlen(s) + pathpos + 2 >= PATH_MAX)
 	return 0;
     while ((pathbuf[pathpos++] = *s++));
     pathbuf[pathpos - 1] = '/';
@@ -1460,7 +1469,7 @@ getfullpath(char *s)
     static char buf[PATH_MAX];
 
     strcpy(buf, pathbuf);
-    strcat(buf, s);
+    strncpy(buf + pathpos, s, PATH_MAX - pathpos - 1);
     return buf;
 }
 
@@ -1511,6 +1520,9 @@ scanner(Complist q)
 		if (!closure || exists(pathbuf))
 		    scanner((q->closure) ? q : q->next);
 		pathbuf[pathpos = oppos] = '\0';
+	    } else if (!*c->str) {
+		if (exists(getfullpath(".")))
+		    insert(dupstring(pathbuf));
 	    } else {
 		/* Last path section.  See if there's a file there. */
 		char *s;
@@ -1583,7 +1595,7 @@ scanner(Complist q)
 #define GF_TOPLEV	2	/* outside (), so ~ ends main match */
 
 static char *pptr;		/* current place in string being matched */
-static Comp tail = 0;
+static Comp tail;
 static int first;		/* are leading dots special? */
 
 /* The main entry point for matching a string str against  *
@@ -1835,6 +1847,7 @@ parsepat(char *str)
 {
     mode = 0;			/* path components present */
     pptr = str;
+    tail = NULL;
     return parsecomplist();
 }
 
@@ -1847,6 +1860,7 @@ parsereg(char *str)
     remnulargs(str);
     mode = 1;			/* no path components */
     pptr = str;
+    tail = NULL;
     return parsecompsw(GF_TOPLEV);
 }
 
