@@ -31,9 +31,23 @@
 
 #include "zsh.h"
 
+/*
+ * Note on curhist: with history active, this points to the
+ * last line actually added to the history list.  With history inactive,
+ * the line does not get added to the list until hend(), if at all.
+ * However, curhist is incremented to reflect the current line anyway.
+ * Thus if the line is not added to the list, curhist must be
+ * decremented in hend().
+ */
+
+/* Bits of histactive variable */
+#define HA_ACTIVE	(1<<0)	/* History mechanism is active */
+#define HA_NOSTORE	(1<<1)	/* Don't store the line when finished */
+#define HA_JUNKED	(1<<2)	/* Last history line was already junked */
+#define HA_NOINC	(1<<3)	/* Don't store, curhist not incremented */
+
 extern int cs, ll;
 
-static Histent curhistent;
 /* Array of word beginnings and endings in current history line. */
 short *chwords;
 /* Max, actual position in chwords.
@@ -61,14 +75,9 @@ hwaddc(int c)
 
 	/* Resize history line if necessary */
 	if (hptr - chline >= hlinesz) {
-	    int flag = 0, oldsiz = hlinesz;
+	    int oldsiz = hlinesz;
 
-	    /* Maybe already linked to a history entry. */
-	    if (curhistent->text == chline)
-		flag = 1;
 	    chline = realloc(chline, hlinesz = oldsiz + 16);
-	    if (flag)
-		curhistent->text = chline;
 	    hptr = chline + oldsiz;
 	}
     }
@@ -188,10 +197,7 @@ getsubsargs(char *subline)
 int
 getargc(Histent ehist)
 {
-  if (ehist == curhistent)
-    return chwordpos ? chwordpos/2-1 : 0;
-  else
-    return ehist->nwords-1;
+    return ehist->nwords ? ehist->nwords-1 : 0;
 }
 
 /* Perform history substitution, returning the next character afterwards. */
@@ -494,56 +500,28 @@ hungetc(int c)
 {
     int doit = 1;
 
-    if (!lexstop) {
-      escape:
+    while (!lexstop) {
+	if (hptr[-1] != (char) c && stophist < 4 &&
+	    hptr > chline + 1 && hptr[-1] == '\n' && hptr[-2] == '\\')
+	    hungetc('\n'), hungetc('\\');
+
 	if (expanding) {
 	    cs--;
 	    ll--;
 	    exlast++;
 	}
-#ifdef DEBUG
-	/* debugging only */
-	if (hptr == chline)
-	    zerr("hungetc attempted at buffer start", NULL, 0);
-	else {
-	    if (hptr[-1] != (char) c)
-		/* Someone probably tried to unget something which is *
-		 * not returned by hgetc.                             */
-		zerr("Oops. You discovered a bug in hungetc", NULL, 0);
-#endif
-	    hptr--;
-	    if (c == bangchar && stophist < 2 &&
-		hptr > chline && hptr[-1] == '\\')
-		qbang = 1;   /* Why do I set it here? See explanation below */
-	    else
-		qbang = 0;
-#ifdef DEBUG
-	}
-#endif
-    }
-    if (doit)
-	inungetc(c);
-    if (c == bangchar && stophist < 2 && qbang) {
-	/* If the character we unget is a bangchar, we have to unget the   *
-	 * backslash as well. The backslash must always be unget from the  *
-	 * strored history word, but if stophist = 1, the backslash should *
-	 * not be put back to the real input to prevent extra backslashes  *
-	 * appear in the input. The problem here is that we could not tell *
-	 * if the current character was really quoted originally if hlastw *
-	 * is NULL (it is NULL at the beginning of a new lex token). But   *
-	 * have to know it if we want correctly move the zle cursor while  *
-	 * expanding history. So we use qbang, which is true, if the last  *
-	 * bangchar was quoted. In some rare cases when more than one      *
-	 * bangs are unget by the lexer, this may give a wrong result, but *
-	 * it can happen only if bangchar is some very unusual character   *
-	 * (unusual means unusual for a bang like a number).  If this bang *
-	 * comes from an alias expansion, no backslash should be really    *
-	 * unget.                                                          */
+	DPUTS(hptr <= chline, "BUG: hungetc attempted at buffer start");
+	hptr--;
+	DPUTS(*hptr != (char) c, "BUG: wrong character in hungetc() ");
+	qbang = (c == bangchar && stophist < 2 &&
+		 hptr > chline && hptr[-1] == '\\');
+	if (doit)
+	    inungetc(c);
+	if (!qbang)
+	    return;
 	doit = !stophist && ((inbufflags & INP_HIST) ||
-			     !(inbufflags & INP_ALIAS));
-	qbang = 0;
+				 !(inbufflags & INP_ALIAS));
 	c = '\\';
-	goto escape;
     }
 }
 
@@ -577,8 +555,10 @@ strinend(void)
 void
 hbegin(void)
 {
+    Histent curhistent;
+
     isfirstln = isfirstch = 1;
-    histremmed = errflag = histdone = spaceflag = 0;
+    errflag = histdone = spaceflag = 0;
     stophist = (!interact || unset(BANGHIST) || unset(SHINSTDIN)) << 1;
     chline = hptr = zcalloc(hlinesz = 16);
     chwords = zalloc((chwordlen = 16)*sizeof(short));
@@ -587,18 +567,13 @@ hbegin(void)
     curhistent = gethistent(curhist);
     if (!curhistent->ftim)
 	curhistent->ftim = time(NULL);
+    histactive = HA_ACTIVE;
     if (interact && isset(SHINSTDIN) && !strin) {
 	attachtty(mypgrp);
-	defev = curhist++;
-	curhistent = gethistent(curhist);
-	zsfree(curhistent->text);
-	if (curhistent->nwords)
-	    zfree(curhistent->words, curhistent->nwords*2*sizeof(short));
-	curhistent->text = chline;
-	curhistent->words = NULL;
-	curhistent->nwords = 0;
+	defev = curhist;
+	curhist++;
     } else
-	histremmed = 1;
+	histactive |= HA_NOINC;
 }
 
 /* say we're done using the history mechanism */
@@ -610,12 +585,14 @@ hend(void)
     int flag, save = 1;
     Histent he;
 
-    if (!chline)
-	return 1;
-    if (!interact || strin || unset(SHINSTDIN)) {
+    DPUTS(!chline, "BUG: chline is NULL in hend()");
+    if (histactive & (HA_NOSTORE|HA_NOINC)) {
 	zfree(chline, hlinesz);
 	zfree(chwords, chwordlen*sizeof(short));
 	chline = NULL;
+	if (!(histactive & HA_NOINC))
+	    curhist--;
+	histactive = 0;
 	return 1;
     }
     flag = histdone;
@@ -652,10 +629,16 @@ hend(void)
 	} else
 	    zsfree(ptr);
     }
-    curhistent->stim = time(NULL);
-    curhistent->ftim = 0L;
-    curhistent->flags = 0;
     if (save) {
+	Histent curhistent = gethistent(curhist);
+	zsfree(curhistent->text);
+	if (curhistent->nwords)
+	    zfree(curhistent->words, curhistent->nwords*2*sizeof(short));
+
+	curhistent->text = ztrdup(chline);
+	curhistent->stim = time(NULL);
+	curhistent->ftim = 0L;
+	curhistent->flags = 0;
 #ifdef DEBUG
 	/* debugging only */
 	if (chwordpos%2) {
@@ -673,17 +656,11 @@ hend(void)
 		   curhistent->nwords*2*sizeof(short));
 	}
     } else
-	remhist();
-    if (chline && !curhistent->text)
-	zfree(chline, hlinesz);
-    if (curhistent->text) {
-	char *s = ztrdup(curhistent->text);
-
-	zfree(curhistent->text, hlinesz);
-	curhistent->text = s;
-    }
+	curhist--;
+    zfree(chline, hlinesz);
     zfree(chwords, chwordlen*sizeof(short));
     chline = NULL;
+    histactive = 0;
     return !(flag & HISTFLAG_NOEXEC || errflag);
 }
 
@@ -693,10 +670,17 @@ hend(void)
 void
 remhist(void)
 {
-    if (!histremmed) {
-	histremmed = 1;
-	curhist--;
-    }
+    if (!(histactive & HA_ACTIVE)) {
+	if (!(histactive & HA_JUNKED)) {
+	    /* make sure this doesn't show up when we do firsthist() */
+	    Histent he = gethistent(curhist);
+	    zsfree(he->text);
+	    he->text = NULL;
+	    histactive |= HA_JUNKED;
+	    curhist--;
+	}
+    } else
+	histactive |= HA_NOSTORE;
 }
 
 /* Gives current expansion word if not last word before chwordpos. */
@@ -712,10 +696,8 @@ hwbegin(int offset)
 	chwordpos--;	/* make sure we're on a word start, not end */
     /* If we're expanding an alias, we should overwrite the expansion
      * in the history.
-     * If we're in a string, we don't need the full history line
-     * and can overwrite also.
      */
-    if (((inbufflags & INP_ALIAS) || strin) && !(inbufflags & INP_HIST))
+    if ((inbufflags & INP_ALIAS) && !(inbufflags & INP_HIST))
 	hwgetword = chwordpos;
     else
 	hwgetword = -1;
@@ -885,7 +867,7 @@ hconsearch(char *str, int *marg)
 	    int pos = s - he->text;
 	    while (t1 < he->nwords && he->words[2*t1] <= pos)
 		t1++;
-	    *marg = t1;
+	    *marg = t1 - 1;
 	    return t0;
 	}
     return -1;
@@ -1059,9 +1041,21 @@ convamps(char *out, char *in, int inlen)
 struct histent *
 quietgethist(int ev)
 {
+    static struct histent storehist;
+
     if (ev < firsthist() || ev > curhist)
 	return NULL;
-    return gethistent(ev);
+    if (ev == curhist && (histactive & HA_ACTIVE)) {
+	/* The current history line has not been stored.  Build it up
+	 * from other variables.
+	 */
+	storehist.text = chline;
+	storehist.nwords = chwordpos/2;
+	storehist.words = chwords;
+
+	return &storehist;
+    } else
+	return gethistent(ev);
 }
 
 /**/
@@ -1091,19 +1085,10 @@ gethist(int ev)
 char *
 getargs(Histent elist, int arg1, int arg2)
 {
-    char *ret;
-    short *words;
-    int pos1, nwords;
+    short *words = elist->words;
+    int pos1, nwords = elist->nwords;
 
-    if (elist == curhistent) {
-      words = chwords;
-      nwords = chwordpos/2;
-    } else {
-      words = elist->words;
-      nwords = elist->nwords;
-    }
-
-    if (arg1 >= nwords || arg2 >= nwords) {
+    if (arg2 < arg1 || arg1 >= nwords || arg2 >= nwords) {
 	/* remember, argN is indexed from 0, nwords is total no. of words */
 	inerrflush();
 	zerr("no such word in event", NULL, 0);
@@ -1111,10 +1096,7 @@ getargs(Histent elist, int arg1, int arg2)
     }
 
     pos1 = words[2*arg1];
-    ret = dupstring(elist->text + pos1);
-    ret[words[2*arg2+1] - pos1] = '\0';
-    
-    return ret;
+    return dupstrpfx(elist->text + pos1, words[2*arg2+1] - pos1);
 }
 
 /**/
@@ -1430,7 +1412,7 @@ savehistfile(char *s, int err, int app)
     else
 	out = fdopen(open(unmeta(s), O_CREAT | O_WRONLY | O_TRUNC, 0600), "w");
     if (out) {
-	for (; ev <= curhist; ev++) {
+	for (; ev <= curhist - !!histactive; ev++) {
 	    ent = gethistent(ev);
 	    if (app & 2) {
 		if (ent->flags & HIST_OLD)
