@@ -109,8 +109,11 @@ loop(int toplevel)
     for (;;) {
 	freeheap();
 	errflag = 0;
-	if (interact && isset(SHINSTDIN))
-	    preprompt();
+	if (isset(SHINSTDIN)) {
+	    setblock_stdin();
+	    if (interact)
+		preprompt();
+	}
 	hbegin();		/* init history mech        */
 	intr();			/* interrupts on            */
 	lexinit();              /* initialize lexical state */
@@ -599,11 +602,13 @@ setupvals(void)
     createparamtable();     /* create paramater hash table             */
 
 #ifdef TIOCGWINSZ
-    adjustwinsize();
+    adjustwinsize(0);
 #else
-    /* Using zero below sets the defaults from termcap */
-    setiparam("COLUMNS", 0);
-    setiparam("LINES", 0);
+    /* columns and lines are normally zero, unless something different *
+     * was inhereted from the environment.  If either of them are zero *
+     * the setiparam calls below set them to the defaults from termcap */
+    setiparam("COLUMNS", columns);
+    setiparam("LINES", lines);
 #endif
 
     /* create hash table for multi-character emacs bindings */
