@@ -156,7 +156,7 @@ update_job(Job jn)
     }
 
     if (shout && !ttyfrozen && !jn->stty_in_env &&
-	job == thisjob && !somestopped)
+	job == thisjob && !somestopped && !(jn->stat & STAT_NOSTTY))
 	gettyinfo(&shttyinfo);
 
     if (isset(MONITOR)) {
@@ -166,7 +166,7 @@ update_job(Job jn)
 	if (mypgrp != pgrp && inforeground &&
 	    (jn->gleader == pgrp || (pgrp > 1 && kill(-pgrp, 0) == -1))) {
 	    attachtty(mypgrp);
-	    adjustwinsize();   /* check window size and adjust if necessary */
+	    adjustwinsize(0);   /* check window size and adjust if necessary */
 	}
     }
 
