@@ -42,7 +42,7 @@ void
 createparamtable(void)
 {
     Param ip, pm;
-    char **old_environ, **envp, **envp2, **sigptr, **t;
+    char **new_environ, **envp, **envp2, **sigptr, **t;
     char buf[50], *str, *iname;
     int num_env;
 
@@ -76,6 +76,10 @@ createparamtable(void)
     noerrs = 1;
 
     HEAPALLOC {
+	int allexp = opts[ALLEXPORT];
+
+	opts[ALLEXPORT] = 0;
+
 	/* Add the standard non-special parameters which have to    *
 	 * be initialized before we copy the environment variables. *
 	 * We don't want to override whatever values the users has  *
@@ -96,35 +100,35 @@ createparamtable(void)
 
 	/* Copy the environment variables we are inheriting to dynamic *
 	 * memory, so we can do mallocs and frees on it.               */
-	old_environ = environ;
-	num_env = arrlen(old_environ);
-	environ = (char **) zalloc(sizeof(char *) * (num_env + 1));
-	*environ = NULL;
+	num_env = arrlen(environ);
+	new_environ = (char **) zalloc(sizeof(char *) * (num_env + 1));
+	*new_environ = NULL;
 
 	/* Now incorporate environment variables we are inheriting *
 	 * into the parameter hash table.                          */
-	for (envp = environ, envp2 = old_environ; *envp2; envp2++) {
+	for (envp = new_environ, envp2 = environ; *envp2; envp2++) {
 	    for (str = *envp2; *str && *str != '='; str++);
 	    if (*str == '=') {
-		iname = NULL;
+		iname = ztrdup(*envp2);
+		str = iname + (str - *envp2);
 		*str = '\0';
-		if (!idigit(**envp2) && isident(*envp2) && !strchr(*envp2, '[')) {
-		    iname = *envp2;
-		    if ((!(pm = (Param) paramtab->getnode(paramtab, iname)) ||
-			 !(pm->flags & PM_DONTIMPORT)) &&
-			(pm = setsparam(iname, metafy(str + 1, -1, META_DUP))) &&
-			!(pm->flags & PM_EXPORTED)) {
-			*str = '=';
-			pm->flags |= PM_EXPORTED;
-			pm->env = *envp++ = ztrdup(*envp2);
-			*envp = NULL;
-			if (pm->flags & PM_SPECIAL)
-			    pm->env = replenv(pm->env, getsparam(pm->nam));
+		if (!idigit(*iname) && isident(iname) && !strchr(iname, '[') &&
+		    !((pm = (Param) paramtab->getnode(paramtab, iname)) &&
+		      (pm->flags & (PM_DONTIMPORT | PM_EXPORTED))) &&
+		    (pm = setsparam(iname, metafy(str + 1, -1, META_DUP)))) {
+		    *str = '=';
+		    pm->flags |= PM_EXPORTED;
+		    if (pm->flags & PM_SPECIAL) {
+			zsfree(iname);
+			iname = mkenvstr(pm->nam, getsparam(pm->nam));
 		    }
-		}
-		*str = '=';
+		    pm->env = *envp++ = iname;
+		    *envp = NULL;
+		} else
+		    zsfree(iname);
 	    }
 	}
+	environ = new_environ;
 
 	pm = (Param) paramtab->getnode(paramtab, "HOME");
 	if (!(pm->flags & PM_EXPORTED)) {
@@ -156,6 +160,7 @@ createparamtable(void)
 	setsparam("ZSH_VERSION", ztrdup(ZSH_VERSION));
 	setaparam("signals", sigptr = zalloc((SIGCOUNT+4) * sizeof(char *)));
 	for (t = sigs; (*sigptr++ = ztrdup(*t++)); );
+	opts[ALLEXPORT] = allexp;
     } LASTALLOC;
 
     noerrs = 0;
@@ -1031,8 +1036,9 @@ setaparam(char *s, char **val)
 	    createparam(t, PM_ARRAY);
 	else if (PM_TYPE(v->pm->flags) != PM_ARRAY &&
 		 !(v->pm->flags & PM_SPECIAL)) {
+	    int uniq = v->pm->flags & PM_UNIQUE;
 	    unsetparam(t);
-	    createparam(t, PM_ARRAY);
+	    createparam(t, PM_ARRAY | uniq);
 	    v = NULL;
 	}
     }
@@ -1234,23 +1240,11 @@ intvarsetfn(Param pm, long x)
 void
 zlevarsetfn(Param pm, long x)
 {
-    if ((long *)pm->data == & columns) {
-	if(x <= 0)
-	    x = tccolumns > 0 ? tccolumns : 80;
-	if (x > 2)
-	    termflags &= ~TERM_NARROW;
-	else
-	    termflags |= TERM_NARROW;
-    } else if ((long *)pm->data == & lines) {
-	if(x <= 0)
-	    x = tclines > 0 ? tclines : 24;
-	if (x > 2)
-	    termflags &= ~TERM_SHORT;
-	else
-	    termflags |= TERM_SHORT;
-    }
+    long *p = (long *)pm->data;
 
-    *((long *)pm->data) = x;
+    *p = x;
+    if (p == &lines || p == &columns)
+	adjustwinsize(2 + (p == &columns));
 }
 
 /* Function to set value of generic special scalar    *
@@ -1330,6 +1324,7 @@ colonarrsetfn(Param pm, char *x)
 
     freearray(*dptr);
     *dptr = x ? colonsplit(x, pm->flags & PM_UNIQUE) : mkarray(NULL);
+    zsfree(x);
     if (pm->ename)
 	arrfixenv(pm->nam, *dptr);
 }
@@ -1424,10 +1419,12 @@ usernamesetfn(Param pm, char *x)
 #endif
 	if(!setgid(pswd->pw_gid) && !setuid(pswd->pw_uid)) {
 	    zsfree(cached_username);
-	    cached_username = ztrdup(pswd->pw_name);
+	    cached_username = x;
 	    cached_uid = pswd->pw_uid;
+	    return;
 	}
     }
+    zsfree(x);
 #endif
 }
 
@@ -1783,15 +1780,16 @@ arrfixenv(char *s, char **t)
     u = zjoin(t, ':');
     len_s = strlen(s);
     pm = (Param) paramtab->getnode(paramtab, s);
-    for (ep = environ; *ep; ep++)
-	if (!strncmp(*ep, s, len_s) && (*ep)[len_s] == '=') {
-	    pm->env = replenv(*ep, u);
-	    return;
-	}
     if (isset(ALLEXPORT))
 	pm->flags |= PM_EXPORTED;
-    if (pm->flags & PM_EXPORTED)
+    if (pm->flags & PM_EXPORTED) {
+	for (ep = environ; *ep; ep++)
+	    if (!strncmp(*ep, s, len_s) && (*ep)[len_s] == '=') {
+		pm->env = replenv(*ep, u);
+		return;
+	    }
 	pm->env = addenv(s, u);
+    }
 }
 
 /* Given *name = "foo", it searchs the environment for string *
