@@ -33,6 +33,7 @@
 
 #define YYERROR  { tok = LEXERR; return NULL; }
 #define YYERRORV { tok = LEXERR; return; }
+#define COND_ERROR(X,Y) do{zerr(X,Y,0);discard_input();YYERROR}while(0)
 
 #define make_list()     allocnode(N_LIST)
 #define make_sublist()  allocnode(N_SUBLIST)
@@ -45,6 +46,23 @@
 #define make_varnode()  allocnode(N_VARASG)
 #define make_cond()     allocnode(N_COND)
 
+static void
+discard_input(void)
+{
+    errflag = 0;
+    if (isnewlin <= 0) {
+	/* Discard remaining stuff after a parse error. */
+	int c;
+
+	hwbegin(0);
+	while ((c = hgetc()) != '\n' && !lexstop);
+	if (c == '\n')
+	    hungetc('\n');
+	hwend();
+    }
+    errflag = 1;
+}
+
 /*
  * event	: ENDINPUT
  *			| SEPER
@@ -103,18 +121,7 @@ par_event(void)
 	    return NULL;
 	}
 	yyerror();
-	errflag = 0;
-	if (isnewlin <= 0) {
-	    /* Discard remaining stuff after a parse error. */
-	    int c;
-
-	    hwbegin(0);
-	    while ((c = hgetc()) != '\n' && !lexstop);
-	    if (c == '\n')
-		hungetc('\n');
-	    hwend();
-	}
-	errflag = 1;
+	discard_input();
 	return NULL;
     } else {
 	l->right = par_event();
@@ -1242,10 +1249,8 @@ par_cond_double(char *a, char *b)
 {
     Cond n = (Cond) make_cond();
 
-    if (a[0] != '-' || !a[1] || a[2]) {
-	zerr("parse error: condition expected: %s", a, 0);
-	return NULL;
-    }
+    if (a[0] != '-' || !a[1] || a[2])
+	COND_ERROR("parse error: condition expected: %s", a);
     n->left = (void *) b;
     n->type = a[1];
     n->ntype = NT_SET(N_COND, 1, NT_STR, NT_STR, 0, 0);
@@ -1284,9 +1289,9 @@ par_cond_triple(char *a, char *b, char *c)
 	if ((t0 = get_cond_num(b + 1)) > -1)
 	    n->type = t0 + COND_NT;
 	else
-	    zerr("unrecognized condition: %s", b, 0);
+	    COND_ERROR("unrecognized condition: %s", b);
     } else
-	zerr("condition expected: %s", b, 0);
+	COND_ERROR("condition expected: %s", b);
     n->left = (void *) a;
     n->right = (void *) c;
     n->ntype = NT_SET(N_COND, 1, NT_STR, NT_STR, 0, 0);
