@@ -420,6 +420,8 @@ strcatsub(char **d, char *pb, char *pe, char *src, int l, char *s, int glbsub)
     return dest;
 }
 
+typedef int (*CompareFn) _((const void *, const void *));
+
 /**/
 int
 strpcmp(const void *a, const void *b)
@@ -704,7 +706,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
     int vunset = 0;
     int spbreak = isset(SHWORDSPLIT) && !ssub && !qt;
     char *val = NULL, **aval = NULL;
-    int fwidth = 0;
+    unsigned int fwidth = 0;
     Value v;
     int flags = 0;
     int flnum = 0;
@@ -1005,7 +1007,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 		fwidth = v->pm->ct ? v->pm->ct : strlen(val);
 		switch (v->pm->flags & (PM_LEFT | PM_RIGHT_B | PM_RIGHT_Z)) {
 		    char *t;
-		    int t0;
+		    unsigned int t0;
 
 		case PM_LEFT:
 		case PM_LEFT | PM_RIGHT_Z:
@@ -1372,12 +1374,7 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 	int i;
 	LinkNode on = n;
 
-	if (!aval[0]) {
-	    if (plan9) {
-		uremnode(l, n);
-		*str = aptr;
-		return n;
-	    }
+	if (!aval[0] && !plan9) {
 	    if (aptr > (char *) getdata(n) &&
 		aptr[-1] == Dnull && *s == Dnull)
 		*--aptr = '\0', s++;
@@ -1388,53 +1385,62 @@ paramsubst(LinkList l, LinkNode n, char **str, int qt, int ssub)
 	    setdata(n, y);
 	    return n;
 	}
-	if (sortit && !copied)
-	    aval = arrdup(aval);
-	if (sortit == 1)
-	    qsort(aval, arrlen(aval), sizeof(char *), strpcmp);
+	if (sortit) {
+	    static CompareFn sortfn[] = {
+		strpcmp, invstrpcmp, cstrpcmp, invcstrpcmp
+	    };
 
-	else if (sortit == 2)
-	    qsort(aval, arrlen(aval), sizeof(char *), invstrpcmp);
-
-	else if (sortit == 3)
-	    qsort(aval, arrlen(aval), sizeof(char *), cstrpcmp);
-
-	else if (sortit)
-	    qsort(aval, arrlen(aval), sizeof(char *), invcstrpcmp);
+	    if (!copied)
+		aval = arrdup(aval);
 
+	    i = arrlen(aval);
+	    if (i && (*aval[i-1] || --i))
+		qsort(aval, i, sizeof(char *), sortfn[sortit-1]);
+	}
 	if (plan9) {
 	    LinkList tl = newlinklist();
+	    LinkNode tn;
 
+	    *--s = Marker;
 	    addlinknode(tl, s);
 	    if (!eval && !stringsubst(tl, firstnode(tl), ssub))
 		return NULL;
 	    *str = aptr;
-	    if (empty(tl)) {
-		uremnode(l, n);
-		return n;
-	    }
-	    i = 0;
-	    while (aval[i]) {
-		LinkNode tn;
-
-		x = aval[i++];
+	    tn = firstnode(tl);
+	    while ((x = *aval++)) {
 		if (prenum || postnum)
 		    x = dopadding(x, prenum, postnum, preone, postone,
 				  premul, postmul);
 		if (eval && parsestr(x))
 		    return NULL;
 		xlen = strlen(x);
-		for (tn = firstnode(tl); tn; incnode(tn)) {
-		    strcatsub(&y, ostr, aptr, x, xlen,
-			      (char *) getdata(tn), globsubst);
+		for (tn = firstnode(tl);
+		     tn && *(y = (char *) getdata(tn)) == Marker;
+		     incnode(tn)) {
+		    strcatsub(&y, ostr, aptr, x, xlen, y + 1, globsubst);
 		    if (qt && !*y && isarr != 2)
 			y = dupstring(nulstring);
-		    if (i == 1)
-			setdata(n, (void *) y);
+		    if (plan9)
+			setdata(n, (void *) y), plan9 = 0;
 		    else
 			insertlinknode(l, n, (void *) y), incnode(n);
 		}
 	    }
+	    for (; tn; incnode(tn)) {
+		y = (char *) getdata(tn);
+		if (*y == Marker)
+		    continue;
+		if (qt && !*y && isarr != 2)
+		    y = dupstring(nulstring);
+		if (plan9)
+		    setdata(n, (void *) y), plan9 = 0;
+		else
+		    insertlinknode(l, n, (void *) y), incnode(n);
+	    }
+	    if (plan9) {
+		uremnode(l, n);
+		return NULL;
+	    }
 	} else {
 	    x = aval[0];
 	    if (prenum || postnum)
