@@ -1110,8 +1110,8 @@ describekeybriefly(void)
 	return;
     trashzle();
     clearflag = (isset(USEZLE) && !termflags &&
-		 (isset(ALWAYSLASTPROMPT) && zmult == 1)) ||
-	(unset(ALWAYSLASTPROMPT) && zmult != 1);
+		 (isset(ALWAYSLASTPROMPT) && !gotmult)) ||
+	(unset(ALWAYSLASTPROMPT) && gotmult);
     printbind(keybuf, (len = strlen(keybuf)) ? len : 1, 0);
     fprintf(shout, " is ");
     if (cmd == z_sendstring) {
@@ -1157,8 +1157,8 @@ whereis(void)
     funcfound = 0;
     trashzle();
     clearflag = (isset(USEZLE) && !termflags &&
-		 (isset(ALWAYSLASTPROMPT) && zmult == 1)) ||
-	(unset(ALWAYSLASTPROMPT) && zmult != 1);
+		 (isset(ALWAYSLASTPROMPT) && !gotmult)) ||
+	(unset(ALWAYSLASTPROMPT) && gotmult);
     if (func == z_selfinsert || func == z_undefinedkey)
 	fprintf(shout, "%s is on many keys", zlecmds[func].name);
     else {
