@@ -763,7 +763,7 @@ putstr(int d)
 void
 tsetcap(int cap, int flag)
 {
-    if (termok && unset(SINGLELINEZLE) && tcstr[cap]) {
+    if (!(termflags & TERM_SHORT) && tcstr[cap]) {
 	switch(flag) {
 	case -1:
 	    tputs(tcstr[cap], 1, putraw);
@@ -803,7 +803,7 @@ tsetcap(int cap, int flag)
 /* get a prompt string */
 
 static char *buf, *bp1, *bl0, *fm, *pmpt;
-static int bracepos, bufspc;
+static int bracepos, bufspc, pmpth;
 
 /**/
 char *
@@ -815,8 +815,9 @@ putprompt(char *fmin, int *lenp, int *wp, int cnt)
 	    *wp = 0;
 	return ztrdup("");
     }
+    pmpth = 1;
 
-    if (!termok && (unset(INTERACTIVE)))
+    if ((termflags & TERM_UNKNOWN) && (unset(INTERACTIVE)))
         init_term();
 
     bracepos = -1;
@@ -841,6 +842,7 @@ putprompt(char *fmin, int *lenp, int *wp, int cnt)
 		*wp = 0;
 		*bp++ = ' ';
 		++*lenp;
+		pmpth++;
 	    }
 	    if (!*wp && *lenp) {
 		addbufspc(1);
@@ -850,6 +852,8 @@ putprompt(char *fmin, int *lenp, int *wp, int cnt)
 	    }
 	}
     }
+    if (pmpt == lpmpt)
+	lppth = pmpth;
 
     return buf;
 }
@@ -1294,9 +1298,12 @@ pputc(char c)
 	    *bp++ = c;
 	    bl0 = bp;
 	    lensb = 0;
+	    pmpth++;
 	    return;
 	}
     }
+    if (bp - bl0 == columns)
+	pmpth++;
     *bp++ = c;
 }
 
