@@ -575,9 +575,10 @@ refreshline(int ln)
 	if (nllen)
 	    strncpy(p1, nl, nllen);
 	memset(p1 + nllen, ' ', winw - nllen);
-	p1[winw] = p1[winw + 1] = '\0';
+	p1[winw] = '\0';
+	p1[winw + 1] = (nllen < winw) ? '\0' : nl[winw + 1];
 	if (ln && nbuf[ln])
-	    strncpy(nl, p1, winw + 2);	/* next time obuf will be up-to-date */
+	    memcpy(nl, p1, winw + 2);	/* next time obuf will be up-to-date */
 	else
 	    nl = p1;		/* don't keep the padding for prompt line */
 	nllen = winw;
