@@ -56,6 +56,12 @@ typedef int LV;
 # define STOUC(X)	((unsigned char)(X))
 #endif
 
+/* Meta together with the character following Meta denotes the character *
+ * which is the exclusive or of 32 and the character following Meta.     *
+ * This is used to represent characters which otherwise has special      *
+ * meaning for zsh.  These are the characters for which the imeta() test *
+ * is true: the null character, and the characters from Meta to Marker.  */
+
 #define Meta		((char) 0x83)
 
 /* Note that the fourth character in DEFAULT_IFS is Meta *
@@ -91,6 +97,9 @@ typedef int LV;
 
 #define INULL(x)	(((x) & 0xfc) == 0x98)
 
+/* Marker used in paramsubst for rc_expand_param */
+#define Marker		((char) 0x9c)
+
 /* chars that need to be quoted if meant literally */
 
 #define SPECCHARS "#$^*()=|{}[]`<>?~;&\n\t \\\'\""
@@ -565,6 +574,8 @@ struct job {
 #define STAT_SUPERJOB	(1<<7)	/* job has a subjob                     */
 #define STAT_SUBJOB	(1<<8)	/* job is a subjob                      */
 #define STAT_CURSH	(1<<9)	/* last command is in current shell     */
+#define STAT_NOSTTY	(1<<10)	/* the tty settings are not inherited   */
+				/* from this job when it exits.         */
 
 #define SP_RUNNING -1		/* fake status for jobs currently running */
 
